-- This script is by Star and the Big Dog. We have also used a little code from Delf's boost mod.
-- If you use any of this, then you must credit all 3 of us.
-- Lots of comments to try and be helpful. Delf and Big Dog's comments have been noted with their name, otherwise they're by Star.

-- CONSOLE COMMANDS
-- This is the global console command "boostmglobal", which toggles boostmomentum on or off for everyone in the server. Defaults to on.
rawset(_G, "CV_boostmglobal", CV_RegisterVar({"boostmglobal", "On", CV_NETVAR, CV_OnOff}))

-- This is the global console command "boostmextreme", which turns on the extreme mode. Defaults to off.
rawset(_G, "CV_boostmextreme", CV_RegisterVar({"boostmextreme", "Off", CV_NETVAR, CV_OnOff}))

-- This is for the player-specific command "boostm", which toggles boostmentum on or off for only the player.
local player_boostm_toggle = function(boostm_active, arg, player)
	local current_bool = player[boostm_active]
	if arg == nil then
		player[boostm_active] = not $1
	elseif arg == "0" or arg == "off" or arg == "false" then
		player[boostm_active] = false
	elseif arg == "1" or arg == "on" or arg == "true" then
		player[boostm_active] = true
	else
		CONS_Printf(player, boostm_active.." should be called with either 'on', 'off', or no argument")
		return
	end
	CONS_Printf(player, boostm_active.." is now "..(player[boostm_active] and "\149deactivated\160. (for yourself)" or "\147activated\160! (for yourself)"))
end

COM_AddCommand("boostm", function(player, arg)
	player_boostm_toggle("boostm", arg, player)
end)

-- INIT THE BIGDOG
-- This is here to make sure everything resets on a map change, as it was causing wall collision issue problems without.
addHook("MapLoad", function(gamemap)
	for player in players.iterate do
		local mobj = player.mo
		player.canboost = false
		player.boosting = false
		player.boostTimerStarted = false
		player.boostStartTime = 0
		player.chargeboost = false
		player.chargeboost2 = false
		player.springCatch = false
		player.springCatch2 = false
		player.globalbooston = false
		player.globalboostoff = false
		player.booston = false
		player.boostoff = false
		player.boostextreme = false
		player.booststandard = false
		player.prevForward = nil
        player.timeOfBlocked = 0
	end
end)

-- Variables for my console command stuff.
local previousGlobalValue
local previousExtremeValue

-- MAIN
-- FUNCTION
local function Boostmentum(player)
-- This line is for compatibility with Xmomentum, which disables its momentum effects when loaded alongside this mod.
	player.hasnomomentum = true 
-- This code is for the global console command boostmglobal to print a message when changed, so players are aware.
    if not (CV_boostmglobal.value == previousGlobalValue) then
    previousGlobalValue = CV_boostmglobal.value
        if CV_boostmglobal.value then
            print("\147Boostmentum is enabled!\160")
        else
            print("\149Boostmentum is disabled.\160")
        end
	end
-- Same thing but for the console command boostmextreme to print when changed.
    if not (CV_boostmextreme.value == previousExtremeValue) then
    previousExtremeValue = CV_boostmextreme.value
        if CV_boostmextreme.value then
            print("\148Boostmentum is set to extreme mode.\160")
        else
            print("\150Boostmentum is set to standard mode.\160")
        end
	end
-- Same code as the last two bunches, but for the sound effects. Only difference is we're using player instead of local.
-- Yes, duplicating the entire code instead of adding different checks is insanely lazy. :)
-- Global sounds.
    if not (CV_boostmglobal.value == player.previousGlobalValue) then
    player.previousGlobalValue = CV_boostmglobal.value
        if CV_boostmglobal.value then
-- This is so we can access the sounds in FREESLOT.LUA, as I like to manage all my sounds in one place.
			player.globalbooston = true
        else
			player.globalboostoff = true
        end
	end
-- Extreme sounds.
    if not (CV_boostmextreme.value == player.previousExtremeValue) then
    player.previousExtremeValue = CV_boostmextreme.value
        if CV_boostmextreme.value then
			player.boostextreme = true
        else
			player.booststandard = true
        end
	end	
-- This code is for the player-only command boostm.
	if not (player.boostm == player.previousBoostValue) then
		player.previousBoostValue = player.boostm
		if player.boostm then
			player.boostoff = true
		else
			player.booston = true
		end
	end
    -- Delf: Prevents the code from running in NiGHTs special stages.
	if (maptol & TOL_NIGHTS) then
		return
	end
	if player.mo and player.mo.valid
		-- Delf: Lets the player boost.
		if not (player.pflags & PF_STASIS)
		and not (player.pflags & PF_FULLSTASIS)
		and not (player.pflags & PF_STARTDASH)
		and not (player.pflags & PF_FINISHED)
		and	player.playerstate ~= PST_DEAD
		and player.powers[pw_carry] == CR_NONE
		and not P_PlayerInPain(player)
-- This is for the console commands.
		and CV_boostmglobal.value
-- This is for the player toggle command.
		and not player.boostm then	
				player.canboost = true
		else
			player.canboost = false
		end
		if player.canboost == nil then
			player.canboost = true
		end

-- BOOST ACTIVATOR
-- MOMENTUM
		-- Delf: Tells the game when the player shall boost, and whenever not.
		local skin = skins[player.mo.skin]
-- This is how we're activating the boost mode by a speed requirement. Require keys to be pressed so it doesn't accelerate by itself.
        if	((player.cmd.forwardmove or player.cmd.sidemove) and not (player.boosting))
-- This specifies the effects to characters that can use spindash.
-- This mod used to have different speed requirements for those characters. I changed that, but kept the separate codes just in case I need it.
            and skins[player.mo.skin].ability2 == CA2_SPINDASH
-- This is our speed requirement.
            and (player.speed >= 70*FRACUNIT)
            and player.canboost
            then player.boosting = true
-- Initliazing this variable for more use later.
			player.springCatch2 = false
        else
-- This second batch of similar code is for non-spindash players. I used skins for the flag, as it doesn't work without it.
        if	((player.cmd.forwardmove or player.cmd.sidemove) and not (player.boosting))
            and not (skins[player.mo.skin].flags & CA2_SPINDASH)
            and (player.speed >= 70*FRACUNIT)
            and player.canboost
            then player.boosting = true
			player.springCatch2 = false
        else
		
-- BOOST ACTIVATOR
-- DURATION
-- This code is for the duration activation of the boost at a much lower speed requirement.
        if	(player.cmd.forwardmove or player.cmd.sidemove)
            and player.canboost
			and (player.speed >= 24*FRACUNIT) then
-- We keep track of the current time here. Then we can add ticrate duration to it below.
			if not player.boostTimerStarted then
				player.boostTimerStarted = true

				--#BIGDOG# Just initializing this varaible
				player.springCatch = false

				player.boostStartTime = player.realtime
				--print("START AT TIME: "+player.boostStartTime)
			else
-- This spawns in a pre-boost trail at halfway 'till boost, to give us an indicator of how much we've charged.
            if (player.realtime >= (player.boostStartTime + (3*TICRATE)))
                and not player.boosting then
				local ghost = P_SpawnGhostMobj(player.mo)
				ghost.fuse = 2
				local ghost = P_SpawnGhostMobj(player.mo)
				ghost.fuse = 2
                --print("REACHED AT TIME: "+player.realtime)
            end
-- This plays a chime at halfway 'till boost. It activates player.chargeboost, which is utilized in the freeslot script for sounds.
-- The sound can be played directly here, but I prefer managing all my sounds in one place.
            if (player.realtime == (player.boostStartTime + (3*TICRATE)))
                and not player.boosting then
				player.chargeboost = true
                --print("REACHED AT TIME: "+player.realtime)
            end
-- Same thing but at 80% of the way until boost.
            if (player.realtime >= (player.boostStartTime + (5*TICRATE)))
                and not player.boosting then
				local ghost = P_SpawnGhostMobj(player.mo)
				ghost.fuse = 3
				local ghost = P_SpawnGhostMobj(player.mo)
				ghost.fuse = 3
                --print("REACHED AT TIME: "+player.realtime)
            end
            if (player.realtime == (player.boostStartTime + (5*TICRATE)))
                and not player.boosting then
				player.chargeboost2 = true
                --print("REACHED AT TIME: "+player.realtime)
            end
-- Actual boost activation.
            if (player.realtime >= (player.boostStartTime + (7*TICRATE)))
                and not player.boosting then
                player.boosting = true
                --print("REACHED AT TIME: "+player.realtime)
            end
		end
		
-- CHARGE GRACE
-- PERIOD
-- Cancelling the charge when interrupted. This code also contains a grace window that allows for a reduced speed, 
-- as well as a time window so that it isn't cancelled by grazing a wall slightly or hitting a spring.
		else			
			--#BIGDOG# Start checking exit conditions only if the boostTimer has started
			if player.boostTimerStarted then
			
				--#BIGDOG# Checks if the player speed has dropped below the threshold required to maintain the boost build up
				if (player.speed <= 20*FRACUNIT)
					or not player.canboost
					or not (player.cmd.forwardmove or player.cmd.sidemove) then				
					--#BIGDOG# If the 'springCatch' has started, it will check how long it has been since the start
					if player.springCatch
					and (player.realtime >= (player.springCatchTime + (TICRATE/5))) then

						--#BIGDOG# If it has been long enough since 'springCatch' started, then it will end the window to regain speed
						--print("SPRING CATCH END")
						player.springCatch = false
					else

						--#BIGDOG# If there is directional input and springCatch has not started, then it will begin the timer for the window to regain speed
						if ((player.cmd.forwardmove or player.cmd.sidemove) and not player.springCatch) then
							--print("SPRING CATCH START "+player.realtime)
							player.springCatch = true
							player.springCatchTime = player.realtime
						end
					end
				
					--#BIGDOG# If 'springCatch' is false, either because there was no directional input or because the timer has run out, then the build up will finish
					if not player.springCatch then
						--print("BOOST TIMER END "+player.realtime)
						player.boostTimerStarted = false
						player.boosting = false
					end
				else

					--#BIGDOG# If the player has regained speed since the 'springCatch' window opened, then it will close the window and continue building up for the boost
					if player.springCatch then
						--print("SPRING CAUGHT SUCCESSFULLY")
						player.springCatch = false
					end
				end
			end
-- This code is the boost charge visual effects from above, used here so the effects won't disappear during the grace period.
        if player.springCatch or player.boostTimerStarted then  
		    if (player.realtime >= (player.boostStartTime + (3*TICRATE)))
                and not player.boosting then
				local ghost = P_SpawnGhostMobj(player.mo)
				ghost.fuse = 2
				local ghost = P_SpawnGhostMobj(player.mo)
				ghost.fuse = 2
            end
            if (player.realtime == (player.boostStartTime + (3*TICRATE)))
                and not player.boosting then
				player.chargeboost = true
            end
            if (player.realtime >= (player.boostStartTime + (5*TICRATE)))
                and not player.boosting then
				local ghost = P_SpawnGhostMobj(player.mo)
				ghost.fuse = 3
				local ghost = P_SpawnGhostMobj(player.mo)
				ghost.fuse = 3
            end
            if (player.realtime == (player.boostStartTime + (5*TICRATE)))
                and not player.boosting then
				player.chargeboost2 = true
            end
		end
            --print("FAILED AT TIME: "+player.realtime)
		end	
		
-- BOOST GRACE
-- PERIOD
-- The same grace period code used right above, but in this case its for the active boost rather than the boost chargeup.
			if player.boosting then

				--#BIGDOG#3-- Initialize variables that check previous tic input amounts
				if (player.prevForward == nil) then
					player.prevForward = player.cmd.forwardmove
				end

				--#BIGDOG#3-- Create variables that check if there has been a large change of direction since the previous tic
				local hasChangedForwardDirection = ( (player.cmd.forwardmove > player.prevForward) and ((player.cmd.forwardmove - player.prevForward) > 40)
						or ( (player.cmd.forwardmove < player.prevForward) and ((player.prevForward - player.cmd.forwardmove) > 40) ) )

				--#BIGDOG#3-- Create variable that checks if overall input is too low
				local inputIsTooLow = ( (player.cmd.forwardmove < 25) and (player.cmd.forwardmove > -25) and (player.cmd.sidemove < 25) and (player.cmd.sidemove > -25) )
				
				--#BIGDOG#3-- Check if conditions are met to see if grace period should be started (tbh not sure why this condition works perfectly)
				if ((player.speed <= 20*FRACUNIT)
				or not player.canboost
				or not hasChangedForwardDirection
				or not (player.cmd.sidemove)
				or not inputIsTooLow) then

					--#BIGDOG#3-- Check if grace period has started and has finished timing...
					if player.springCatch2
					and (player.realtime >= (player.springCatchTime2 + (TICRATE/5))) then

						--#BIGDOG#3-- ...if so, end the grace period
						player.springCatch2 = false
					else

						--#BIGDOG#3-- ...if not, check if there's enough input, and if direction hasn't changed enough, to start the grace period
						if (not hasChangedForwardDirection or (player.cmd.sidemove)) and not inputIsTooLow then

							--#BIGDOG#3-- ...if so, start the grace period if not started already
							if not player.springCatch2 then
								player.springCatch2 = true
								player.springCatchTime2 = player.realtime
								--print("BOOST GRACE START @ "+player.realtime +", Changed fw dir? "+hasChangedForwardDirection+", Sidemove? "+(player.cmd.sidemove == true)+", Input Too Low? "+inputIsTooLow)
							end
						else

							--#BIGDOG#3-- ...if not, then the grace period won't be started due to insufficient input, or because direction has changed
							player.boosting = false 
							player.boostTimerStarted = false
							--print("NO BOOST GRACE PERIOD, SO END @ "+player.realtime+", Changed fw dir? "+hasChangedForwardDirection+", Sidemove? "+(player.cmd.sidemove == true)+", Input Too Low? "+inputIsTooLow)
						
							if inputIsTooLow then
								--print("Input was too low @ "+player.realtime+", ForwardMove: "+player.cmd.forwardmove+", Sidemove: "+player.cmd.sidemove)
							end
						end
					end
				else
					--print("BOOST GRACE PERIOD AVOIDED. Speed too low? "+(player.speed <= 20*FRACUNIT)+", CanBoost? "+player.canboost+", Changed fw dir? "+hasChangedForwardDirection+", sidemove? "+(player.cmd.sidemove == true)+", Input Too Low? "+inputIsTooLow)
				end

				--#BIGDOG#3-- Check if grace period has finished/not been started
				if not player.springCatch2 then
					
					--#BIGDOG#3-- End the boost if grace period has finished
					player.boosting = false 
					player.boostTimerStarted = false
					--print("BOOST GRACE END ")
				else

					--#BIGDOG#3-- ???
					if player.springCatch2 then
						player.springCatch2 = false
					end
				end

				--#BIGDOG#3-- While boosting, set the 'previous tic input amounts' to the current tic's amounts. Next tic, on lines 196 and 199, these 'previous tic' values will be used
				if player.boosting then
					player.prevForward = player.cmd.forwardmove
				end
			end

			--#BIGDOG#3-- If not boosting, set the 'previous tic input amounts' to 'nil' so that they get re-initialized on the next boost on lines 188 and 191
			if not player.boosting then
				player.prevForward = nil
			end
		end
	end

-- BOOSTING
-- EFFECTS
		if player.boosting then
-- This is a hotfix to ghost trails bugging out during boosts.
			player.boostTimerStarted = false
-- This is the code for the trail that appears behind you. I'm using 3 sets - one colored white, to get the desired visual effect.
			local ghost = P_SpawnGhostMobj(player.mo)
			ghost.fuse = 4
			local ghost = P_SpawnGhostMobj(player.mo)
			ghost.fuse = 4
			local ghost = P_SpawnGhostMobj(player.mo)
			ghost.fuse = 4
			ghost.blendmode = AST_ADD
			ghost.colorized = true
			ghost.color = player.mo.color
		end		
		-- Delf: Dermines how fast shall the player boost.
-- Two separate speed modes, for regular and extreme.
		player.boostspeed = 16 * player.normalspeed / 14
		player.extremeboostspeed = 21 * player.normalspeed / 14		
		-- Delf: Stores the player's speed angle.
		if player.speed > player.normalspeed then
			player.momangle2 = R_PointToAngle2(0,0,player.mo.momx,player.mo.momy)
		else
			player.momangle2 = player.mo.angle
		end
		-- Delf: What happens if the player boosts.
		if player.boosting then
--			if not player.cmd.forwardmove then
--				player.cmd.forwardmove = 50 --Intentionally not in PreThinkFrame
--			end
			player.drawangle = player.momangle2
-- This fixes a water running bug where you get stuck when half-submerged.
			if player.mo and player.mo.eflags&MFE_UNDERWATER then
				player.charflags = $|SF_NOSKID|SF_CANBUSTWALLS
				
				if not (skins[player.mo.skin].flags & SF_RUNONWATER) then
					player.charflags = $ & ~SF_RUNONWATER
				end
			else
				player.charflags = $|SF_NOSKID|SF_RUNONWATER|SF_CANBUSTWALLS
			end
-- Here's where the boost speed triggers, which depends on extreme mode or not.
			if player.speed < (player.boostspeed) 
				and not CV_boostmextreme.value then
				P_Thrust(player.mo, player.momangle2, player.boostspeed - player.speed)
			end
			if player.speed < (player.extremeboostspeed) 
				and CV_boostmextreme.value then
				P_Thrust(player.mo, player.momangle2, player.extremeboostspeed - player.speed)
			end
		else 
		-- Delf: What happens if the player isn't boosting. Mostly resetting values.
			if not (skins[player.mo.skin].flags & SF_NOSKID) then
				player.charflags = $ & ~SF_NOSKID
			end
			if not (skins[player.mo.skin].flags & SF_RUNONWATER) then
				player.charflags = $ & ~SF_RUNONWATER
			end
			if not (skins[player.mo.skin].flags & SF_CANBUSTWALLS ) then
				player.charflags = $ & ~SF_CANBUSTWALLS
			end
		end
	end
end

-- NECESSARY
-- HOOKS
addHook("PlayerThink", Boostmentum)

addHook("PlayerCanDamage", function(player) -- Player kills enemies while boosting
    if player.boosting then
        return true
    end
end)

-- This hook makes walls stop your boost and works well with the grace period, by not interrupting your charge/ boost from tiny bumps.
addHook("MobjMoveBlocked", function(mo) -- You shall not boost if stopped by walls.
    local player = mo.player

    if(player.timeOfBlocked == nil)
        player.timeOfBlocked = player.realtime-(1)
    end

    if(player.realtime > (player.timeOfBlocked + (1)))
        player.timeOfBlocked = player.realtime
        --print("START TIMING WALL TOUCH: "+player.realtime)
    end

    if(player.realtime == (player.timeOfBlocked + (1))) then
        player.canboost = false
        player.boosting = false
		player.boostTimerStarted = false
		player.springCatch = false
        --print("TOUCHED WALL TOO LONG: "+player.realtime)
    end
end, MT_PLAYER)

addHook("MobjMoveCollide", function(mo, spike) -- Spike breaking by ARJr.
    if mo.valid
    and (mo.type == MT_PLAYER)
	and mo.player.boosting
    and spike.type == MT_SPIKE
		if (spike.z >= (mo.z - 16*FRACUNIT)) and (spike.z <= (mo.z + mo.height + 16*FRACUNIT))
			P_KillMobj(spike, mo, mo)
		elseif (spike.flags2 & MF2_OBJECTFLIP)
		and (spike.z <= (mo.z - 16*FRACUNIT)) and (spike.z >= (mo.z + mo.height+ 16*FRACUNIT))
			P_KillMobj(spike, mo, mo)
        end
    end
end)